------------
-- funcs.sql
--
--      Define access functions, etc for the Veil demo database.
--
--      Copyright (c) 2005, 2006 Marc Munro
--      Author:  Marc Munro
--	License: BSD
--
-- $Id: funcs.sql,v 1.2 2006/10/04 15:50:09 bloodnok Exp $
------------

\echo - Creating access control functions...

-- Veil Variables:
--   Shared:
--     privs_range Range The range of privilege_id in the privileges table
--     roles_range Range The range of role_id in the roles table
--     role_privs Bitmap Array Array of privilege bitmaps indexed by role_id
--     det_types_range Range The range of detail_type_id in detail_types
--     det_types_privs Int4Array Mapping of detail_type_id to the privilege
--       required to access details of that type
--     
--   Session:
--     person_id Int4 The id of the currently connected person
--     global_context Bitmap The global privileges of the connected person
--     project_context Bitmap Hash Hash of privilege bitmaps, indexed by
--       project_id, giving the connected person's project privileges.
--
-- Security Contexts:
--   Global
--     Privileges are identified in variable global_context
--   Personal
--     Privileges are identified in role_privs[11002].
--   Project
--     A relational context.  Privileges are identified in
--     project_context[project_id].

-- Replacement for default version of veil_init.
-- Initialise and load all shared variables.
--
\echo - ...veil_init...
create or replace
function veil_init(bool) returns bool as '
declare
    doing_reset alias for $1;
    exists_privs_range bool;
    exists_roles_range bool;
    exists_role_privs  bool;
    exists_det_types_range bool;
    exists_det_types_privs bool;
    init_reqd bool;
    dummy     bool;
    dummy2    bool;
    dummyint  int4;
    _count    int4;
begin
    -- Declare all shared variables.

    select into exists_privs_range, exists_roles_range,
		exists_role_privs, exists_det_types_range,
		exists_det_types_privs
           veil_share(''privs_range''), veil_share(''roles_range''),
	   veil_share(''role_privs''), veil_share(''det_types_range''),
	   veil_share(''det_types_privs'');

    init_reqd = not (exists_privs_range and exists_role_privs and
		     exists_role_privs and exists_det_types_range and
		     exists_det_types_privs);

    if init_reqd or doing_reset then
    	-- Load ranges for privs and roles.
    	select into dummyint
    	       veil_init_range(''roles_range'', min(role_id), max(role_id))
    	from   vdemo_owner.roles;

    	select into dummyint
    	       veil_init_range(''privs_range'', min(privilege_id), 
    				  max(privilege_id))
    	from   vdemo_owner.privileges;

	-- Load range for detail_types
    	select into dummyint
    	       veil_init_range(''det_types_range'', 
	       		       min(detail_type_id), max(detail_type_id))
    	from   vdemo_owner.detail_types;

	-- Initialise array of required privs for detail_types
    	select into dummy
    	       veil_init_int4array(''det_types_privs'', ''det_types_range'');
    	select into _count
    	       count(veil_int4array_set(''det_types_privs'', 
	       		                detail_type_id, required_privilege_id))
    	from   vdemo_owner.detail_types;

    	-- Initialise role_privs bitmap_array
    	select into dummy
    		   veil_init_bitmap_array(''role_privs'', ''roles_range'', 
    					 ''privs_range'');

    	-- Populate role_privs bitmap_array
    	select into _count
    		   count(veil_bitmap_array_setbit(''role_privs'', 
    						 role_id, privilege_id))
    	from   vdemo_owner.role_privileges;

    end if;

    -- Declare important session variables, so that we do not 
    -- get odd, undefined variable, error messages.
    select into dummyint, dummy, dummy2
	   veil_int4_set(''person_id'',  null),
	   veil_init_bitmap(''global_context'', ''privs_range''),
	   veil_init_bitmap_hash(''project_context'', ''privs_range'');

    return true;		        
end;
' language plpgsql
stable 
security definer;

revoke execute on function veil_init(bool) from public;


\echo - ...connect_person...
create or replace
function connect_person(int4) returns bool as '
declare
    _person_id alias for $1;
    dummy    int4;
    _connect bool;
    proj_roles record;
    last_proj int4;
    first_rec bool;
begin
    -- In reality this function would require some authentication token such 
    -- as a password.  This is just a dumb demo version.

    select into _connect disconnect_person();

    -- Test whether provided person exists.  This is where we would, in a
    -- real version, do proper authentication.
    select into dummy 1
    from   vdemo_owner.persons
    where  person_id = _person_id;

    if found then
	-- The person exists and passes authentication

	-- From the persons roles set the global_context bitmap.
	select into dummy
	       count(veil_union_from_bitmap_array(''global_context'',
						  ''role_privs'', role_id))
	from   vdemo_owner.person_roles
	where  person_id = _person_id;

	-- Check that user has can_connect privilege
	select into _connect
	    veil_bitmap_testbit(''global_context'', 10100);

	if not _connect then
	    select into _connect disconnect_person();
	    return false;
	end if;


	-- From the persons assignments set the project_context bitmap hash.
	select into dummy
	       count(veil_union_into_bitmap_hash(''project_context'',
			project_id::text,
			veil_bitmap_from_array(''scratch_bitmap'',
					       ''role_privs'', role_id)))
	from   vdemo_owner.assignments
	where  person_id = _person_id;

	-- Finally, record the person_id for the connection.
	select into dummy veil_int4_set(''person_id'', _person_id);

	return true;
    else
	return false;
    end if;

end;
' language plpgsql
stable 
security definer;

\echo - ...disconnect_person...
create or replace
function disconnect_person() returns bool as '
declare
    dummy int4;
    dummy2 bool;
    dummy3 bool;
begin
    -- Clear session bitmaps, and reset the person_id
    select into dummy, dummy2, dummy3
	   veil_int4_set(''person_id'', null),
	   veil_init_bitmap(''global_context'', ''privs_range''),
	   veil_init_bitmap_hash(''project_context'', ''privs_range'');
    return false;
end;
' language plpgsql
stable 
security definer;


\echo - ...i_have_global_priv...
create or replace
function i_have_global_priv(int4) returns bool as '
declare
    priv_id alias for $1;
    connection_id int4;
    result bool;
begin
    select into connection_id, result 
           veil_int4_get(''person_id''), 
           veil_bitmap_testbit(''global_context'', priv_id);
    if connection_id is null then
	return false;
    else
        return result;
    end if;
end;
' language plpgsql
stable 
security definer;


\echo - ...i_have_personal_priv...
create or replace
function i_have_personal_priv(int4, int4) returns bool as '
declare
    priv_id alias for $1;
    person_id alias for $2;
    connection_id int4;
    result bool;
begin
    select into connection_id, result 
           veil_int4_get(''person_id''), 
           veil_bitmap_testbit(''global_context'', priv_id);
    if connection_id is null then
	-- No-one is connected
	return false;
    else
	if result then
	    -- We have the required privilege in global context.  No need
	    -- to check any further
	    return true;
	else
	    if person_id = connection_id then
		-- We are connected as the owner of this record.  Check
		-- whether we have the required privilege in personal 
		-- context.
		select into result
	               veil_bitmap_array_testbit(''role_privs'', 
					        11002, priv_id);
		return result;
	    else
		-- We have no personal context rights to this record
		return false;
	    end if;
	end if;
    end if;
end;
' language plpgsql
stable 
security definer;


\echo - ...i_have_project_priv...
create or replace
function i_have_project_priv(int4, int4) returns bool as '
declare
    priv_id alias for $1;
    project_id alias for $2;
    connection_id int4;
    result bool;
begin
    select into connection_id, result 
           veil_int4_get(''person_id''), 
           veil_bitmap_testbit(''global_context'', priv_id);
    if connection_id is null then
	-- No-one is connected
	return false;
    else
	if result then
	    -- We have the required privilege in global context.  No need
	    -- to check any further
	    return true;
	else
	    select into result
	    	veil_bitmap_hash_testbit(''project_context'', 
				         project_id, priv_id);
	    return result;
	end if;
    end if;
end;
' language plpgsql
stable 
security definer;


\echo - ...i_have_proj_or_pers_priv...
create or replace
function i_have_proj_or_pers_priv(int4, int4, int4) returns bool as '
declare
    priv_id alias for $1;
    project_id alias for $2;
    person_id alias for $3;
    connection_id int4;
    result bool;
begin
    select into connection_id, result 
           veil_int4_get(''person_id''), 
           veil_bitmap_testbit(''global_context'', priv_id);
    if connection_id is null then
	-- No-one is connected
	return false;
    else
	if result then
	    -- We have the required privilege in global context.  No need
	    -- to check any further
	    return true;
	else
	    if person_id = connection_id then
		-- We are connected as the owner of this record.  Check
		-- whether we have the required privilege in personal 
		-- context.
		select into result
	               veil_bitmap_array_testbit(''role_privs'', 
					        11002, priv_id);
		return result;
	    end if;
	    select into result
	    	veil_bitmap_hash_testbit(''project_context'', 
				         project_id, priv_id);
	    return result;
		-- We have no personal context rights to this record
		-- so check project context
		return false;
	end if;
    end if;
end;
' language plpgsql
stable 
security definer;

\echo - ...i_have_project_detail_priv...
create or replace
function i_have_project_detail_priv(int4, int4) returns bool as '
declare
    result bool;
    detail_id alias for $1;
    proj_id alias for $2;
begin
    select into result 
        i_have_project_priv(veil_int4array_get(''det_types_privs'', detail_id),
			    proj_id);
    return result;
end;
' language plpgsql
stable 
security definer;

\echo - ...i_have_person_detail_priv...
create or replace
function i_have_person_detail_priv(int4, int4) returns bool as '
declare
    result bool;
    detail_id alias for $1;
    person_id alias for $2;
begin
    select into result 
       i_have_personal_priv(veil_int4array_get(''det_types_privs'', detail_id),
			    person_id);
    return result;
end;
' language plpgsql
stable 
security definer;

/*
Loading role_roles

function populate_rr_privs(in_role_id)
begin
    if in_role_id in loaded_roles then 
	-- Stop when we get a cyclic dependency
	return
    end if;

    loaded_roles |= in_role_id;

    for role_id in select owned_role_id from role_roles 
		   where owning_role_id = in_role_id
    loop
	populate_rr_privs(role_id);
	role_privs[in_role_id] |= role_privs[role_id];
    end loop;
end;


*/
