/**
 * @file   veil_shmem_pre82.inc
 * \code
 *     Author:       Marc Munro
 *     Copyright (c) 2006 Marc Munro
 *     License:      BSD
 * $Id: veil_shmem_pre82.inc,v 1.2 2007/07/31 22:18:27 bloodnok Exp $
 * \endcode
 * @brief  
 * This file handles shared memory for version of postgres prior to 
 * postgres 8.2.  It is conditionally included into veil_shmem.c
 */

/**
 * Limits the number of shared variables that can be created in each
 * context.
 */
#define SHARED_HASH_ELEMS     32

/**
 * shared_meminfo provides access to the ShmemCtl structure allocated in
 * the first chunk of context 0.
 */
static ShmemCtl *shared_meminfo = NULL;
static LWLockId  VeilLWLock;

/**
 * Whether the current backend is in the process of switching contexts.
 * If so, it will be setting up the non-current context in readiness for
 * making it available to all other backends.
 */
static bool      prepared_for_switch = false;


/** 
 * Return the index of the other context from the one supplied.
 * 
 * @param x the context for which we want the other one.
 * 
 * @return the opposite context to that of x.
 */
#define OTHER_CONTEXT(x) 	(x ? 0: 1)


/** 
 * Create/attach to the shared hash identified by hashname.  Return a
 * pointer to an HTAB that references the shared hash.
 * 
 * @param hashname 
 * 
 * @return Pointer to HTAB referencing the shared hash.
 */
static HTAB *
create_shared_hash(const char *hashname)
{
	HASHCTL hashctl;
	HTAB   *result;
	char   *db_hashname;

	// Add the current database oid into the hashname so that it is
	// distinct from the shared hash for other databases in the cluster.
	db_hashname = (char *) vl_malloc(HASH_KEYLEN);
	(void) snprintf(db_hashname, HASH_KEYLEN - 1, "%s_%u", 
					hashname, MyDatabaseId);
	hashctl.keysize = HASH_KEYLEN;
	hashctl.entrysize = sizeof(VarEntry);

	LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);
	result = ShmemInitHash(db_hashname, SHARED_HASH_ELEMS,
						   SHARED_HASH_ELEMS, &hashctl, HASH_ELEM);
	LWLockRelease(VeilLWLock);
	pfree(db_hashname);
	return result;
}

/** 
 * Return reference to the HTAB for the shared hash associated with
 * context 0.
 * 
 * @return Pointer to HTAB referencing shared hash for context 0.
 */
static HTAB *
get_hash0()
{
	static HTAB *hash0 = NULL;

    if (!hash0) {
		hash0 = create_shared_hash("VEIL_SHARED1");
	}
	return hash0;
}

/** 
 * Return reference to the HTAB for the shared hash associated with
 * context 1.
 * 
 * @return Pointer to HTAB referencing shared hash for context 1.
 */
static HTAB *
get_hash1()
{
	static HTAB *hash1 = NULL;

    if (!hash1) {
		hash1 = create_shared_hash("VEIL_SHARED2");
	}

	return hash1;
}

/** 
 * Allocate a new chunk of shared memory large enough to satisfy
 * request, recording the amount allocated in allocated.  The new
 * chunk is returned as a ::MemChunk structure, suitably initialised.
 * This is a low level function that does no locking.  If we are unable
 * to allocate sufficient memory an ERROR is raised.
 * 
 * @param request The amount of memory requested.
 * @param allocated This is set to the amount of memory actually
 * allocated.
 * 
 * @return Pointer to chunk of memory.
 */
static MemChunk *
new_chunk(size_t  request,
		  size_t *allocated)
{
	int chunks = (int) ((request + (sizeof(MemChunk) + 1)) / CHUNK_SIZE) + 1;
	size_t actual = (size_t) chunks * CHUNK_SIZE;
	MemChunk *chunk = (MemChunk *) ShmemAlloc(actual);

	if (chunk) {
		chunk->next_chunk = NULL;
		chunk->next = sizeof(MemChunk);
		chunk->limit = actual;
		*allocated = actual;
	}
	else {
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("Out of shared memory")));
	}
	return chunk;
}

/** 
 * Do the leg-work for allocating a chunk of memory.  Do no
 * error-handling, no allocation of new chunks of memory, and no
 * locking.
 * 
 * @param chunk Pointer to the first chunk within which we should search
 * for a suitably sized piece of free memory.
 * @param size The amount of memory we are requesting.
 * 
 * @return Pointer to the piece of memory allocated, or NULL if no
 *sufficiently large piece is avaialable
 */
static void *
do_shmalloc(MemChunk *chunk, size_t size)
{
	size_t amount = (size_t) MAXALIGN(size);
	void  *result;

	while (true) {
		if ((amount + chunk->next) <= chunk->limit) {
			result = chunk + chunk->next;
			chunk->next += amount;
			return result;
		}

		chunk = chunk->next_chunk;
		if (!chunk) {
			return NULL;
		}
	}
}

/** 
 * Figure out which LWLock to use.  In versions prior to 8.1 attempting
 * to assign an LWLock dynamically fails as there are never any free
 * LWLocks available. 
 */
static LWLockId
get_veil_lwlock()
{
	LWLockId result;
#if PG_MAJOR_VERSION == 7
	result = MMCacheLock;
#elif PG_MAJOR_VERSION == 8
	#if PG_MINOR_VERSION == 0
		result = MMCacheLock;
	#else
		// result = LWLockAssign();
		result = NumFixedLWLocks;
	#endif
#else
	result = LWLockAssign();
#endif
	return result;
}

/** 
 * Attach to, creating and initialising as necessary, the shared memory
 * control structure.  Record this for the session in shared_meminfo.
 */
static void
shmalloc_init()
{
	VarEntry *var;
	MemChunk *chunk;
    bool      found;
	size_t    allocated;

	// As this variable only exists in hash0 and shared memory etc is
	// not yet set up, we cannot use define_shared_variable and so we
	// use hash_search directly.

	var = (VarEntry *) hash_search(get_hash0(), (void *) "VEIL_SHMEMCTL",
								   HASH_ENTER, &found);

	if (found) {
		shared_meminfo = (ShmemCtl *) var->obj;
		while (!shared_meminfo->initialised) {
			// Wait for shared_meminfo to be intialised.  This loop
			// is to prevent an unlikely race-condition.
			;
		}
		VeilLWLock = shared_meminfo->veil_lwlock;
	}
	else {
		VeilLWLock = get_veil_lwlock();

		LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);
		// Create initial memory chunk.
		chunk = new_chunk(1, &allocated);
		
		// Allocate and initialise the ShmemCtl structure from our 
		// initial chunk.
		shared_meminfo = do_shmalloc(chunk, sizeof(ShmemCtl));
		shared_meminfo->type = OBJ_SHMEMCTL;
		shared_meminfo->veil_lwlock = VeilLWLock;
		shared_meminfo->current_context = 0;
		shared_meminfo->total_allocated[0] = allocated;
		shared_meminfo->total_allocated[1] = 0;
		shared_meminfo->switching = false;
		shared_meminfo->context[0] = chunk;
		shared_meminfo->context[1] = NULL;
		shared_meminfo->xid[0] = GetCurrentTransactionId();
		shared_meminfo->xid[1] = shared_meminfo->xid[0];
		shared_meminfo->initialised = true;
		var->obj = (Object *) shared_meminfo;
		var->shared = true;
		LWLockRelease(VeilLWLock);
	}
}

/** 
 * Return the id (index) of the current context for this session 
 * 
 * @return The current context id
 */
extern int
get_cur_context_id()
{
	static initialised = false;
	int    context;

	if (!initialised) {
		shmalloc_init();
		initialised = true;
	}
		
	context = shared_meminfo->current_context;
	if (prepared_for_switch) {
		context = OTHER_CONTEXT(context);
	}
	else {
		// Check for the default context being for a later transaction
		// than current and, if so, use the other one.
		if (TransactionIdPrecedes(GetCurrentTransactionId(), 
								  shared_meminfo->xid[context]))
		{
			context = OTHER_CONTEXT(context);
		}
	}
    return context;
}

/** 
 * Return the first chunk of shared memory for the current context.
 * 
 * @return The first chunk of shared memory for the current context. 
 */
static MemChunk *
get_cur_context()
{
    int context;
	context = get_cur_context_id();
    return shared_meminfo->context[context];
}

/** 
 * Return the shared hash for the current context.
 * 
 * @return Pointer to the HTAB for the current context's shared hash.
 */
HTAB *
vl_get_shared_hash()
{
	int context;
	HTAB *hash;
	static initialised = false;

	if (!initialised) {
		(void) get_cur_context();  // Ensure shared memory is set up.
		initialised = true;
	}

	// context = shared_meminfo->current_context;xxxx
	context = get_cur_context_id();

	if (context == 0) {
		hash = get_hash0();
	}
	else {
		hash = get_hash1();
	}
	
	return hash;
}

static void
clear_hash(HTAB *hash)
{
	static HASH_SEQ_STATUS status;
	VarEntry *var;

	hash_seq_init(&status, hash);
	while (var = hash_seq_search(&status)) {
		if (strncmp("VEIL_SHMEMCTL", var->key, strlen("VEIL_SHMEMCTL")) != 0) {
			(void) hash_search(hash, var->key, HASH_REMOVE, NULL);
		}
	}
}

/** 
 * Prepare for a switch to the alternate context.  Switching will
 * only be allowed if there are no transactions that may still be using
 * the context to which we are switching, and there is no other
 * process attempting the switch.
 * 
 * @return true if the switch preparation was successful.
 */
bool
vl_prepare_context_switch()
{
	int   context_curidx;
	int   context_newidx;
	HTAB *hash0 = get_hash0(); // We must not attempt to create hashes on
	HTAB *hash1 = get_hash1(); // the fly below as they also acquire the lock

	(void) get_cur_context();  // Ensure shared memory is set up

	LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);

	if (shared_meminfo->switching) {
		// Another process is performing the switch
		LWLockRelease(VeilLWLock);
		return false;
	}

	shared_meminfo->switching = true;

	// We have claimed the switch.  If we decide that we cannot proceed,
	// we will return it to its previous state.

	context_curidx = shared_meminfo->current_context;
	context_newidx = OTHER_CONTEXT(context_curidx);
	if (shared_meminfo->context[context_newidx]) {
		// The alternate chunk list has been used before, so we have to
		// try to clear it.
		TransactionId oldest_xid;
		MemChunk *chunk;

		oldest_xid = GetOldestXmin(false);
		if (TransactionIdPrecedes(oldest_xid, 
								  shared_meminfo->xid[context_curidx])) 
		{
			// There is a transaction running that precedes the time of
			// the last chunklist switch.  That transaction may still be
			// using the chunk to which we wish to switch.  We cannot
			// allow the switch.
			shared_meminfo->switching = false;
			LWLockRelease(VeilLWLock);
			return false;
		}
		else {

			// It looks like we can safely make the switch.  Reset all
			// the chunks in the new list, and make it the current list
			// for this session only.
			chunk = shared_meminfo->context[context_newidx];

			while (chunk) {
				chunk->next = sizeof(MemChunk);
				chunk = chunk->next_chunk;
			}
			// If we are switching to context 0, reset the next field of
			// the first chunk to leave space for the ShmemCtl struct.
			if (context_newidx == 0) {
				chunk = shared_meminfo->context[context_newidx];
				chunk->next += sizeof(ShmemCtl);
			}

			if (context_newidx == 0) {
				clear_hash(hash0);
			}
			else {
				clear_hash(hash1);
			}
		}
	}
	else {
		// No-one has used this memory set before.  Create a new chunk.
		shared_meminfo->context[context_newidx] = 
			new_chunk(1, &shared_meminfo->total_allocated[context_newidx]);
	}
	
	LWLockRelease(VeilLWLock);
	prepared_for_switch = true;
	return true;
}

/** 
 * Complete the context switch started by vl_prepare_context_switch().
 * Raise an ERROR if the context switch cannot be completed.
 * 
 * @return true if the context switch is successfully completed.
 */
bool
vl_complete_context_switch()
{
	int  context_curidx;
	int  context_newidx;

    if (!prepared_for_switch) {
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("failed to complete context switch"),
				 errdetail("Not prepared for switch - "
						   "invalid state for operation")));
	}

	LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);
	context_curidx = shared_meminfo->current_context;
	context_newidx = OTHER_CONTEXT(context_curidx);

	if (!shared_meminfo->switching) {
		// We do not claim to be switching.  We should.
		LWLockRelease(VeilLWLock);

		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("failed to complete context switch"),
				 errdetail("Session does not have switching set to true- "
						   "invalid state for operation")));
	}

	shared_meminfo->switching = false;
	shared_meminfo->current_context = context_newidx;
	shared_meminfo->xid[context_newidx] = GetCurrentTransactionId();
	LWLockRelease(VeilLWLock);
	prepared_for_switch = false;
	return true;
}

/** 
 * In desparation, if we are unable to complete a context switch, we
 * should use this function.  This will force a context switch and then
 * cause a PANIC, resetting the database.
 */
void
vl_force_context_switch()
{
	int  context_curidx;
	int  context_newidx;

	(void) get_cur_context();

	LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);

	// We don't bother with the switch - everything we do is within
	// VeilLWLock and is, in any case, an act of desparation.

	context_curidx = shared_meminfo->current_context;
	context_newidx = OTHER_CONTEXT(context_curidx);

	if (shared_meminfo->context[context_newidx]) {
		// The alternate chunk list has been used before, so we have to
		// clear it.
		MemChunk *chunk;

		chunk = shared_meminfo->context[context_newidx];

		while (chunk) {
			chunk->next = sizeof(MemChunk);
			chunk = chunk->next_chunk;
		}
	}
	else {
		// No-one has used this chunklist before.  Create a new chunk.
		shared_meminfo->context[context_newidx] = 
			new_chunk(1, &shared_meminfo->total_allocated[context_newidx]);
	}
	
	shared_meminfo->switching = false;
	shared_meminfo->current_context = context_newidx;
	shared_meminfo->xid[context_newidx] = GetCurrentTransactionId();
	shared_meminfo->context[1] = NULL;
	shared_meminfo->xid[0] = GetCurrentTransactionId();
	LWLockRelease(VeilLWLock);
	prepared_for_switch = false;
}

/** 
 * Dynamically allocate a piece of shared memory from the current context. 
 * 
 * @param size The size of the requested piece of memory.
 * 
 * @return Pointer to dynamically allocated memory.
 */
void *
vl_shmalloc(size_t size)
{
	MemChunk *chunk;
	void     *result;
	int       context;

	//context = shared_meminfo->current_context;xxxx
	context = get_cur_context_id();
	chunk = get_cur_context();

	LWLockAcquire(VeilLWLock, LW_EXCLUSIVE);
	result = do_shmalloc(chunk, size);
	if (!result) {
		while (chunk->next_chunk) {
			chunk = chunk->next_chunk;
		}
		chunk->next_chunk = 
			new_chunk(size, &shared_meminfo->total_allocated[context]);
		result = do_shmalloc(chunk, size);
	}
	LWLockRelease(VeilLWLock);
	return result;
}

/** 
 * Free a piece of shared memory within the current context.  Currently
 * this does nothing as implementation of freeing of shared memory has
 * been deferred.
 * 
 * @param mem Pointer to the memory to be freed.
 * 
 */
void
vl_free(void *mem)
{
	return;
}


